from .lib import execnet
from .connection import Connection, FakeRemoteLogger


class _RSync(execnet.RSync):
    """
    Inherits from ``execnet.RSync`` so that we can log nicely with the user
    logger instance (if any) back with the ``_report_send_file`` method
    """

    def __init__(self, sourcedir, callback=None, verbose=True, logger=None):
        self.logger = logger
        super(_RSync, self).__init__(sourcedir, callback, verbose)

    def _report_send_file(self, gateway, modified_rel_path):
        if self._verbose:
            self.logger.info("syncing file: %s" % modified_rel_path)


def rsync(hosts, source, destination, logger=None, sudo=False):
    """
    Grabs the hosts (or single host), creates the connection object for each
    and set the rsync execnet engine to push the files.

    It assumes that all of the destinations for the different hosts is the
    same. This deviates from what execnet does because it has the flexibility
    to push to different locations.
    """
    logger = logger or FakeRemoteLogger()
    sync = _RSync(source, logger=logger)

    # setup_targets
    if not isinstance(hosts, list):
        hosts = [hosts]

    for host in hosts:
        conn = Connection(
            host,
            logger,
            sudo,
        )
        sync.add_target(conn.gateway, destination)

    return sync.send()



import pytest

from ceph_deploy.cli import get_parser
from ceph_deploy.tests.util import assert_too_few_arguments


class TestParserUninstall(object):

    def setup(self):
        self.parser = get_parser()

    def test_uninstall_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('uninstall --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy uninstall' in out
        assert 'positional arguments:' in out
        assert 'optional arguments:' in out

    def test_uninstall_host_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('uninstall'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_uninstall_one_host(self):
        args = self.parser.parse_args('uninstall host1'.split())
        assert args.host == ['host1']

    def test_uninstall_multiple_hosts(self):
        hostnames = ['host1', 'host2', 'host3']
        args = self.parser.parse_args(['uninstall'] + hostnames)
        assert frozenset(args.host) == frozenset(hostnames)



import os
import subprocess

import pytest
from mock import patch, MagicMock, Mock

from ceph_deploy.cli import _main as main
from ceph_deploy.hosts import remotes
from ceph_deploy.tests.directory import directory


def test_bad_no_conf(tmpdir, cli):
    with pytest.raises(cli.Failed) as err:
        with cli(
            args=['ceph-deploy', 'admin', 'host1'],
            stderr=subprocess.PIPE,
            ) as p:
            result = p.stderr.read().decode('utf-8')
    assert 'No such file or directory: \'ceph.conf\'' in result
    assert err.value.status == 1


def test_bad_no_key(tmpdir, cli):
    with tmpdir.join('ceph.conf').open('w'):
        pass
    with pytest.raises(cli.Failed) as err:
        with cli(
            args=['ceph-deploy', 'admin', 'host1'],
            stderr=subprocess.PIPE,
            ) as p:
            result = p.stderr.read().decode('utf-8')
    assert 'ceph.client.admin.keyring not found' in result
    assert err.value.status == 1


def test_write_keyring(tmpdir):
    with tmpdir.join('ceph.conf').open('w'):
        pass
    with tmpdir.join('ceph.client.admin.keyring').open('wb'):
        pass

    etc_ceph = os.path.join(str(tmpdir), 'etc', 'ceph')
    os.makedirs(etc_ceph)

    distro = MagicMock()
    distro.conn = MagicMock()
    remotes.write_file.__defaults__ = (0o644, str(tmpdir), -1, -1)
    distro.conn.remote_module = remotes
    distro.conn.remote_module.write_conf = Mock()

    with patch('ceph_deploy.admin.hosts'):
        with patch('ceph_deploy.admin.hosts.get', MagicMock(return_value=distro)):
            with directory(str(tmpdir)):
                main(args=['admin', 'host1'])

    keyring_file = os.path.join(etc_ceph, 'ceph.client.admin.keyring')
    assert os.path.exists(keyring_file)

    file_mode = oct(os.stat(keyring_file).st_mode & 0o777)
    assert file_mode == oct(0o600)



from . import mon  # noqa
from .install import install, mirror_install, repo_install  # noqa
from .uninstall import uninstall  # noqa
from ceph_deploy.util import pkg_managers
from ceph_deploy.util.system import is_systemd, is_upstart

# Allow to set some information about this distro
#

distro = None
release = None
codename = None


def choose_init(module):
    """
    Select a init system

    Returns the name of a init system (upstart, sysvinit ...).
    """
    # Upstart checks first because when installing ceph, the
    # `/lib/systemd/system/ceph.target` file may be created, fooling this
    # detection mechanism.
    if is_upstart(module.conn):
        return 'upstart'

    if is_systemd(module.conn) or module.conn.remote_module.path_exists(
            "/lib/systemd/system/ceph.target"):
        return 'systemd'

    return 'sysvinit'


def get_packager(module):
    return pkg_managers.Apt(module)



"""
A utility module that can host utilities that will be used by more than
one type of distro and not common to all of them
"""
from ceph_deploy.util import pkg_managers


def install_yum_priorities(distro, _yum=None):
    """
    EPEL started packaging Ceph so we need to make sure that the ceph.repo we
    install has a higher priority than the EPEL repo so that when installing
    Ceph it will come from the repo file we create.

    The name of the package changed back and forth (!) since CentOS 4:

    From the CentOS wiki::

        Note: This plugin has carried at least two differing names over time.
        It is named yum-priorities on CentOS-5 but was named
        yum-plugin-priorities on CentOS-4. CentOS-6 has reverted to
        yum-plugin-priorities.

    :params _yum: Used for testing, so we can inject a fake yum
    """
    yum = _yum or pkg_managers.yum
    package_name = 'yum-plugin-priorities'

    if distro.normalized_name == 'centos':
        if distro.release[0] != '6':
            package_name = 'yum-priorities'
    yum(distro.conn, package_name)





ceph_repo = """[ceph]
name=Ceph packages for $basearch
baseurl={repo_url}/$basearch
enabled=1
gpgcheck={gpgcheck}
priority=1
type=rpm-md
gpgkey={gpg_url}

[ceph-noarch]
name=Ceph noarch packages
baseurl={repo_url}/noarch
enabled=1
gpgcheck={gpgcheck}
priority=1
type=rpm-md
gpgkey={gpg_url}

[ceph-source]
name=Ceph source packages
baseurl={repo_url}/SRPMS
enabled=0
gpgcheck={gpgcheck}
type=rpm-md
gpgkey={gpg_url}
"""

zypper_repo = """[ceph]
name=Ceph packages
type=rpm-md
baseurl={repo_url}
gpgcheck={gpgcheck}
gpgkey={gpg_url}
enabled=1
"""


def custom_repo(**kw):
    """
    Repo files need special care in that a whole line should not be present
    if there is no value for it. Because we were using `format()` we could
    not conditionally add a line for a repo file. So the end result would
    contain a key with a missing value (say if we were passing `None`).

    For example, it could look like::

        [ceph repo]
        name= ceph repo
        proxy=
        gpgcheck=

    Which breaks. This function allows us to conditionally add lines,
    preserving an order and be more careful.

    Previously, and for historical purposes, this is how the template used
    to look::

        custom_repo =
        [{repo_name}]
        name={name}
        baseurl={baseurl}
        enabled={enabled}
        gpgcheck={gpgcheck}
        type={_type}
        gpgkey={gpgkey}
        proxy={proxy}

    """
    lines = []

    # by using tuples (vs a dict) we preserve the order of what we want to
    # return, like starting with a [repo name]
    tmpl = (
        ('reponame', '[%s]'),
        ('name', 'name=%s'),
        ('baseurl', 'baseurl=%s'),
        ('enabled', 'enabled=%s'),
        ('gpgcheck', 'gpgcheck=%s'),
        ('_type', 'type=%s'),
        ('gpgkey', 'gpgkey=%s'),
        ('proxy', 'proxy=%s'),
        ('priority', 'priority=%s'),
    )

    for line in tmpl:
        tmpl_key, tmpl_value = line  # key values from tmpl

        # ensure that there is an actual value (not None nor empty string)
        if tmpl_key in kw and kw.get(tmpl_key) not in (None, ''):
            lines.append(tmpl_value % kw.get(tmpl_key))

    return '\n'.join(lines)



import errno
import logging
import os
from ceph_deploy import hosts, exc
from ceph_deploy.lib import remoto


LOG = logging.getLogger(__name__)


def distro_is_supported(distro_name):
    """
    An enforcer of supported distros that can differ from what ceph-deploy
    supports.
    """
    supported = ['centos', 'redhat', 'ubuntu', 'debian']
    if distro_name in supported:
        return True
    return False


def connect(args):
    for hostname in args.hosts:
        distro = hosts.get(hostname, username=args.username)
        if not distro_is_supported(distro.normalized_name):
            raise exc.UnsupportedPlatform(
                distro.distro_name,
                distro.codename,
                distro.release
            )

        LOG.info(
            'Distro info: %s %s %s',
            distro.name,
            distro.release,
            distro.codename
        )
        LOG.info('assuming that a repository with Calamari packages is already configured.')
        LOG.info('Refer to the docs for examples (http://ceph.com/ceph-deploy/docs/conf.html)')

        rlogger = logging.getLogger(hostname)

        # Emplace minion config prior to installation so that it is present
        # when the minion first starts.
        minion_config_dir = os.path.join('/etc/salt/', 'minion.d')
        minion_config_file = os.path.join(minion_config_dir, 'calamari.conf')

        rlogger.debug('creating config dir: %s' % minion_config_dir)
        distro.conn.remote_module.makedir(minion_config_dir, [errno.EEXIST])

        rlogger.debug(
            'creating the calamari salt config: %s' % minion_config_file
        )
        distro.conn.remote_module.write_file(
            minion_config_file,
            ('master: %s\n' % args.master).encode('utf-8')
        )

        distro.packager.install('salt-minion')
        distro.packager.install('diamond')

        # redhat/centos need to get the service started
        if distro.normalized_name in ['redhat', 'centos']:
            remoto.process.run(
                distro.conn,
                ['chkconfig', 'salt-minion', 'on']
            )

            remoto.process.run(
                distro.conn,
                ['service', 'salt-minion', 'start']
            )

        distro.conn.exit()


def calamari(args):
    if args.subcommand == 'connect':
        connect(args)


def make(parser):
    """
    Install and configure Calamari nodes. Assumes that a repository with
    Calamari packages is already configured. Refer to the docs for examples
    (http://ceph.com/ceph-deploy/docs/conf.html)
    """
    calamari_parser = parser.add_subparsers(dest='subcommand')
    calamari_parser.required = True

    calamari_connect = calamari_parser.add_parser(
        'connect',
        help='Configure host(s) to connect to Calamari master'
    )
    calamari_connect.add_argument(
        '--master',
        nargs='?',
        metavar='MASTER SERVER',
        help="The domain for the Calamari master server"
    )
    calamari_connect.add_argument(
        'hosts',
        nargs='+',
    )

    parser.set_defaults(
        func=calamari,
    )



import logging
import os.path

from ceph_deploy import exc
from ceph_deploy import conf
from ceph_deploy.cliutil import priority
from ceph_deploy import hosts

LOG = logging.getLogger(__name__)


def config_push(args):
    conf_data = conf.ceph.load_raw(args)

    errors = 0
    for hostname in args.client:
        LOG.debug('Pushing config to %s', hostname)
        try:
            distro = hosts.get(hostname, username=args.username)

            distro.conn.remote_module.write_conf(
                args.cluster,
                conf_data,
                args.overwrite_conf,
            )

            distro.conn.exit()

        except RuntimeError as e:
            LOG.error(e)
            errors += 1

    if errors:
        raise exc.GenericError('Failed to config %d hosts' % errors)


def config_pull(args):

    topath = '{cluster}.conf'.format(cluster=args.cluster)
    frompath = '/etc/ceph/{cluster}.conf'.format(cluster=args.cluster)

    errors = 0
    for hostname in args.client:
        try:
            LOG.debug('Checking %s for %s', hostname, frompath)
            distro = hosts.get(hostname, username=args.username)
            conf_file_contents = distro.conn.remote_module.get_file(frompath)

            if conf_file_contents is not None:
                LOG.debug('Got %s from %s', frompath, hostname)
                if os.path.exists(topath):
                    with open(topath, 'rb') as f:
                        existing = f.read()
                        if existing != conf_file_contents and not args.overwrite_conf:
                            LOG.error('local config file %s exists with different content; use --overwrite-conf to overwrite' % topath)
                            raise

                with open(topath, 'wb') as f:
                    f.write(conf_file_contents)
                return
            distro.conn.exit()
            LOG.debug('Empty or missing %s on %s', frompath, hostname)
        except:
            LOG.error('Unable to pull %s from %s', frompath, hostname)
        finally:
            errors += 1

    raise exc.GenericError('Failed to fetch config from %d hosts' % errors)


def config(args):
    if args.subcommand == 'push':
        config_push(args)
    elif args.subcommand == 'pull':
        config_pull(args)
    else:
        LOG.error('subcommand %s not implemented', args.subcommand)


@priority(70)
def make(parser):
    """
    Copy ceph.conf to/from remote host(s)
    """
    config_parser = parser.add_subparsers(dest='subcommand')
    config_parser.required = True

    config_push = config_parser.add_parser(
        'push',
        help='push Ceph config file to one or more remote hosts'
        )
    config_push.add_argument(
        'client',
        metavar='HOST',
        nargs='+',
        help='host(s) to push the config file to',
        )

    config_pull = config_parser.add_parser(
        'pull',
        help='pull Ceph config file from one or more remote hosts'
        )
    config_pull.add_argument(
        'client',
        metavar='HOST',
        nargs='+',
        help='host(s) to pull the config file from',
        )
    parser.set_defaults(
        func=config,
        )



import pytest

from ceph_deploy.cli import get_parser
from ceph_deploy.tests.util import assert_too_few_arguments


class TestParserMDS(object):

    def setup(self):
        self.parser = get_parser()

    def test_mds_help(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('mds --help'.split())
        out, err = capsys.readouterr()
        assert 'usage: ceph-deploy mds' in out
        assert 'positional arguments:' in out
        assert 'optional arguments:' in out

    def test_mds_create_host_required(self, capsys):
        with pytest.raises(SystemExit):
            self.parser.parse_args('mds create'.split())
        out, err = capsys.readouterr()
        assert_too_few_arguments(err)

    def test_mds_create_one_host(self):
        args = self.parser.parse_args('mds create host1'.split())
        assert args.mds[0][0] == 'host1'

    def test_mds_create_multiple_hosts(self):
        hostnames = ['host1', 'host2', 'host3']
        args = self.parser.parse_args(['mds', 'create'] + hostnames)
        # args.mds is a list of tuples, and tuple[0] is the hostname
        hosts = [x[0] for x in args.mds]
        assert frozenset(hosts) == frozenset(hostnames)



import socket
import sys
from .lib import execnet


#
# Connection Object
#

class Connection(object):

    def __init__(self, hostname, logger=None, sudo=False, threads=1, eager=True,
                 detect_sudo=False, interpreter=None):
        self.sudo = sudo
        self.hostname = hostname
        self.logger = logger or FakeRemoteLogger()
        self.remote_module = None
        self.channel = None
        self.global_timeout = None  # wait for ever

        self.interpreter = interpreter or 'python%s' % sys.version_info[0]

        if eager:
            try:
                if detect_sudo:
                    self.sudo = self._detect_sudo()
                self.gateway = self._make_gateway(hostname)
            except OSError:
                self.logger.error(
                    "Can't communicate with remote host, possibly because "
                    "%s is not installed there" % self.interpreter
                )
                raise

    def _make_gateway(self, hostname):
        gateway = execnet.makegateway(
            self._make_connection_string(hostname)
        )
        gateway.reconfigure(py2str_as_py3str=False, py3str_as_py2str=False)
        return gateway

    def _detect_sudo(self, _execnet=None):
        """
        ``sudo`` detection has to create a different connection to the remote
        host so that we can reliably ensure that ``getuser()`` will return the
        right information.

        After getting the user info it closes the connection and returns
        a boolean
        """
        exc = _execnet or execnet
        gw = exc.makegateway(
            self._make_connection_string(self.hostname, use_sudo=False)
        )

        channel = gw.remote_exec(
            'import getpass; channel.send(getpass.getuser())'
        )

        result = channel.receive()
        gw.exit()

        if result == 'root':
            return False
        self.logger.debug('connection detected need for sudo')
        return True

    def _make_connection_string(self, hostname, _needs_ssh=None, use_sudo=None):
        _needs_ssh = _needs_ssh or needs_ssh
        interpreter = self.interpreter
        if use_sudo is not None:
            if use_sudo:
                interpreter = 'sudo ' + interpreter
        elif self.sudo:
            interpreter = 'sudo ' + interpreter
        if _needs_ssh(hostname):
            return 'ssh=%s//python=%s' % (hostname, interpreter)
        return 'popen//python=%s' % interpreter

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.exit()
        return False

    def execute(self, function, **kw):
        return self.gateway.remote_exec(function, **kw)

    def exit(self):
        self.gateway.exit()

    def import_module(self, module):
        self.remote_module = ModuleExecute(self.gateway, module, self.logger)
        return self.remote_module


class ModuleExecute(object):

    def __init__(self, gateway, module, logger=None):
        self.channel = gateway.remote_exec(module)
        self.module = module
        self.logger = logger

    def __getattr__(self, name):
        if not hasattr(self.module, name):
            msg = "module %s does not have attribute %s" % (str(self.module), name)
            raise AttributeError(msg)
        docstring = self._get_func_doc(getattr(self.module, name))

        def wrapper(*args):
            arguments = self._convert_args(args)
            if docstring:
                self.logger.debug(docstring)
            self.channel.send("%s(%s)" % (name, arguments))
            try:
                return self.channel.receive()
            except Exception as error:
                # Error will come as a string of a traceback, remove everything
                # up to the actual exception since we do get garbage otherwise
                # that points to non-existent lines in the compiled code
                for tb_line in reversed(str(error).split('\n')):
                    if tb_line:
                        exc_line = tb_line
                        break
                raise RuntimeError(exc_line)

        return wrapper

    def _get_func_doc(self, func):
        try:
            return getattr(func, 'func_doc').strip()
        except AttributeError:
            return ''

    def _convert_args(self, args):
        if args:
            if len(args) > 1:
                arguments = str(args).rstrip(')').lstrip('(')
            else:
                arguments = str(args).rstrip(',)').lstrip('(')
        else:
            arguments = ''
        return arguments


#
# FIXME this is getting ridiculous
#

class FakeRemoteLogger:

    def error(self, *a, **kw):
        pass

    def debug(self, *a, **kw):
        pass

    def info(self, *a, **kw):
        pass

    def warning(self, *a, **kw):
        pass


def needs_ssh(hostname, _socket=None):
    """
    Obtains remote hostname of the socket and cuts off the domain part
    of its FQDN.
    """
    _socket = _socket or socket
    fqdn = _socket.getfqdn()
    if hostname == fqdn:
        return False
    local_hostname = _socket.gethostname()
    local_short_hostname = local_hostname.split('.')[0]
    if local_hostname == hostname or local_short_hostname == hostname:
        return False
    return True
